<?php
 
namespace Truevo\PaymentGateway\Controller\Adminhtml\Base;
 
use Magento\Backend\App\Action;
use Magento\Backend\App\Action\Context;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Model\Order;
use Truevo\PaymentGateway\Helper\Data as CoreHelper;
use Truevo\PaymentGateway\Model\TruevoAPI;
 
class TruevoAction extends Action
{
    /**
     * How many times to attempt to fetch a transaction status before aborting.
     */
    const MAX_QUERY_STATUS_ATTEMPTS = 5;

    /**
     * How many attempts to get the transaction status have already been made.
     *
     * @var integer
     */
    private $attempts = 0;

    /** 
     * The repository holding all the orders.
     * 
     * @var OrderRepositoryInterface 
     */
    protected $_orderRepository;

    /**
     * The Data helper file
     *
     * @var CoreHelper
     */
    private $_coreHelper;

    /**
     * The order being interacted with.
     *
     * @var Order
     */
    protected $_order;

    /**
     * The handler class for the Truevo API
     *
     * @var TruevoAPI
     */
    protected $_truevoApi;

    /**
     * The transaction details sent back from the initial Truevo payment request.
     *
     * @var array
     */
    protected $_transactionDetails;
 
    /**
     * Constructor
     *
     * @param Context $context
     * @param OrderRepositoryInterface $orderRepository
     * @param CoreHelper $coreHelper
     */
    public function __construct(
        Context $context,
        OrderRepositoryInterface $orderRepository,
        CoreHelper $coreHelper
    )
    {
        parent::__construct($context);
        $this->_orderRepository = $orderRepository;
        $this->_coreHelper      = $coreHelper;
    }
 
    /**
     * The function called when the route gets hit.
     *
     * @return void
     */
    public function execute()
    {
        $this->setOrder();

        $this->setTruevoApi();

        $this->setTransactionDetails();
    }

    /**
     * Set the order variable based off the order_id in the request.
     *
     * @return void
     */
    private function setOrder()
    {
        $orderId = $this->getRequest()->getParam('order_id');
        $this->_order = $this->_orderRepository->get($orderId);

        if (!$this->_order) {
            throw new Exception('Could not find order with the following ID: ' . $orderId);
        }
    }

    /**
     * Set the Truevo API handler class.
     *
     * @return void
     */
    private function setTruevoApi()
    {
        $this->_truevoApi = new TruevoAPI($this->_order);
    }

    /**
     * Set the Truevo API handler class.
     *
     * @return void
     */
    private function setTransactionDetails()
    {
        $this->_transactionDetails = $this->_truevoApi->getTransactionResponseDetails();
    }

    /**
     * Get the url of the show order page via the order ID.
     *
     * @param integer $orderId
     * @return string
     */
    protected function redirectToOrderView(int $orderId) 
    {
        $orderUrl = $this->getUrl('sales/order/view', ['order_id' => $orderId]);

        $this->_redirect($orderUrl);
    }

    /**
     * Build an array from the outcome of a response.
     *
     * @param string $action CAPTURE | QUERY STATUS | REFUND | REVERSAL
     * @param object $response
     * @param array $extraData
     * @return array
     */
    protected function buildLogArray($action, $response, $extraData = [])
    {
        // Rebuild the additional data to a proper key value pair array
        $additionalData = array();
        foreach ($response->additionalData as $keyValuePair) {
            $additionalData[$keyValuePair->key] = $keyValuePair->value;
        }
        
        $data = [
            'action'             => $action,
            'status'             => $response->outcome->status,
            'transaction_status' => $response->outcome->transactionStatus,
            'response_code'      => $response->outcome->responseCode ?? null,
            'reason_description' => $response->outcome->reasonDescription ?? null,
            'amount'             => isset($response->amount->totalAmount) ? $response->amount->currencyAlphaCode . number_format($response->amount->totalAmount, 2) : null,
            'number_of_pieces'   => $additionalData['number_of_pieces'] ?? null,
            'reason_for_refund'  => $additionalData['reason_for_refund'] ?? null
        ];

        $return = array_merge($data, $extraData);

        return array_filter($return);
    }

    /**
     * Add a message to the customer thread of an order.
     *
     * @param array|string $content
     * @return void
     */
    protected function logToOrder($content)
    {
        if (is_array($content)) {
            $message = '';
            foreach ($content as $key => $value) {
                $name = str_replace('_', ' ', $key);
                $message .= ucwords($name) . ": " . $value . "<br/>";
            }
        } else if (is_string($content)) {
            $message = $content;
        } else {
            return;
        }

        if (!$this->_order->addStatusHistoryComment($message)->setIsCustomerNotified(false)->save()) {
            $this->_coreHelper->loggingTruevo('Failed to capture response from cURL request for order #' . $this->_order->getIncrementId());
        }
    }

    /**
     * Poll the status for a transaction unique identifier until either a 
     * "COMPLETE" or "DECLINED" is returned or the request times out.
     *
     * @param string $transactionId
     * @return string "COMPLETE", "DECLINED", or "PENDING"
     */
    protected function pollingTransactionStatus($transactionId)
    {
        $status = "PENDING";
        $isComplete = false;

        do {
            sleep(1); // Wait a second before querying the status
            $this->attempts += 1; // Increment the total number of attempts

            $response = $this->_truevoApi->getStatus($transactionId);

            if ($response) {
                $status = $response->outcome->transactionStatus;

                // Don't log pending requests unless it is the final one
                if ($status !== "PENDING" || $this->attempts === self::MAX_QUERY_STATUS_ATTEMPTS) {
                    $log = $this->buildLogArray('QUERY STATUS', $response);
                    $this->logToOrder($log);
                }

                $isComplete = $status !== "PENDING"; 
            }
        } while ($this->attempts < self::MAX_QUERY_STATUS_ATTEMPTS && !$isComplete);

        return $status;
    }
}