<?php

namespace Truevo\PaymentGateway\Controller\Adminhtml\Refund;

use Magento\Backend\App\Action;
use Magento\Backend\App\Action\Context;
use Magento\Framework\App\Action\HttpGetActionInterface;
use Magento\Framework\View\Result\Page;
use Magento\Framework\View\Result\PageFactory;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Model\Order;
use Truevo\PaymentGateway\Controller\Adminhtml\Base\TruevoAction;
use Truevo\PaymentGateway\Helper\Data as CoreHelper;

class Index extends TruevoAction
{
    /**
     * @var PageFactory
     */
    protected $_resultPageFactory;

    /**
     * Index constructor.
     *
     * @param Context $context
     * @param PageFactory $resultPageFactory
     */
    public function __construct(
        Context $context,
        PageFactory $resultPageFactory,
        OrderRepositoryInterface $orderRepository,
        CoreHelper $coreHelper
    ) {
        parent::__construct($context, $orderRepository, $coreHelper);

        $this->_resultPageFactory = $resultPageFactory;
    }

    /**
     * Load the page defined in view/adminhtml/layout/truevo_refund_index.xml
     *
     * @return Page
     */
    public function execute()
    {
        // Setup the class variables
        parent::execute();

        // Check if this is a POST request from the form
        if ($this->getRequest()->getPostValue()) {
            $isRedirecting = $this->handlePost();

            if ($isRedirecting) {
                return;
            }
        }

        // If not a POST request, or the POST failed, return the page for the refund form
        return $this->handleGet();
    }

    /**
     * Validate the form values for a POST request and send through the request to the Truevo API.
     *
     * @return bool Return true if redirecting, otherwise return false if staying on the refund form page.
     */
    private function handlePost()
    {
        if (!$this->isValid()) {
            $this->messageManager->addError(__('There were some validation errors in the form.'));
            return false;
        }
        
        // Call Truevo API
        $response = $this->_truevoApi->refund(
            (float) $this->getRequest()->getParam('refund_amount'), 
            (int) $this->getRequest()->getParam('number_of_pieces'), 
            $this->getRequest()->getParam('reason_for_refund')
        );

        // Responses
        if ($response) {
            if ($response->outcome->status === "RECEIVED") {
                $transactionUniqueIdentifier = $response->transactionUniqueIdentifier;

                $log = $this->buildLogArray('REVERSAL', $response);
                $this->logToOrder($log);

                switch ( $this->pollingTransactionStatus($transactionUniqueIdentifier) ) {
                    case 'COMPLETE':
                        // If the entire amount has been refunded, cancel the order
                        if ((float) $this->getRequest()->getParam('refund_amount') === (float) $this->getRequest()->getParam('available_for_refund')) {
                            $this->_order->setData('state', Order::STATE_CANCELED);
                            $this->_order->setStatus(Order::STATE_CANCELED);
                        }

                        $this->_order->save();

                        // Successfuly refunded payment
                        $this->messageManager->addSuccess(__('Truevo refund request successfully sent.'));
                        break;
                    case 'DECLINED':
                        $this->messageManager->addError(__('Truevo payment declined the refund.'));
                        break;
                    default: 
                        // Payment is still pending
                        $this->messageManager->addNotice(__('Refund is still pending.'));
                        break;
                }

                $this->redirectToOrderView($this->_order->getId());
                return true;
            }
        }

        // Error refunding payment
        $this->messageManager->addError(__('There was an error requesting a refund for the Truevo payment.'));
        $this->redirectToOrderView($this->_order->getId());
        return false;
    }

    /**
     * Build out the refund form for a GET request.
     *
     * @return PageFactory
     */
    private function handleGet()
    {
        $resultPage = $this->_resultPageFactory->create();
        $resultPage->getConfig()->getTitle()->prepend(__('#' . $this->_order->getIncrementId() . ' / Refund / Orders / Operations'));

        /** @var Template $block */
        $block = $resultPage->getLayout()->getBlock('truevo.refund.form');

        $block->setData('min_refund_amount', 0);
        $block->setData('max_refund_amount', (float) $this->getRequest()->getParam('available_for_refund'));
        $block->setData('min_number_of_pieces', 0);
        $block->setData('max_number_of_pieces', $this->_order->getTotalItemCount());

        return $resultPage;
    }

    /**
     * Checks if the input required is valid and return true if it is, otherwise return false.
     *
     * @return boolean
     */
    private function isValid()
    {
        $details = $this->_truevoApi->getDetails();

        if ($details && $details->paymentDetails) {
            $availableForRefund = $details->paymentDetails->availableForRefund;

            $refundAmount   = (float) $this->getRequest()->getParam('refund_amount');
            $numberOfPieces = (int) $this->getRequest()->getParam('number_of_pieces');
            $refundReason   = $this->getRequest()->getParam('reason_for_refund');

            // If all conditions are met, true will be returned
            return (
                $refundAmount == number_format($refundAmount, 2) 
                && $refundAmount >= 0 
                && $refundAmount <= $availableForRefund 
                && $numberOfPieces >= 0 
                && $numberOfPieces <= $this->_order->getTotalItemCount()
                && strlen($refundReason) <= 150
            );
        }

        return false;
    }
}
