<?php

namespace Truevo\PaymentGateway\Model;

use Exception;
use Magento\Checkout\Model\Session;
use Magento\Customer\Model\Session as CustomerSession;
use Magento\Framework\Api\AttributeValueFactory;
use Magento\Framework\Api\ExtensionAttributesFactory;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Locale\ResolverInterface;
use Magento\Framework\Model\Context;
use Magento\Framework\Registry;
use Magento\Framework\Session\Generic;
use Magento\Framework\UrlInterface;
use Magento\Payment\Helper\Data;
use Magento\Payment\Model\Method\AbstractMethod;
use Magento\Payment\Model\Method\Logger;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Vault\Api\PaymentTokenManagementInterface;
use Truevo\PaymentGateway\Helper\Data as CoreHelper;
use Truevo\PaymentGateway\Model\TruevoConfig;

class PaymentMethod extends AbstractMethod
{
    /**
     * @var TruevoConfig
     */
    protected $_truevoConfig;

    /**
     * @var CustomerSession
     */
    protected $_customerSession;

    /**
     * @var PaymentTokenManagementInterface
     */
    protected $_tokenManagement;

    /**
     * @var CoreHelper
     */
    private $_coreHelper;

    /**
     * @var Args
     */
    protected $_args;

    /**
     * Constructor
     *
     * @param Context $context
     * @param Registry $registry
     * @param ExtensionAttributesFactory $extensionFactory
     * @param AttributeValueFactory $customAttributeFactory
     * @param StoreManagerInterface $storeManager
     * @param Data $paymentData
     * @param ScopeConfigInterface $scopeConfig
     * @param Logger $logger
     * @param Session $checkoutSession
     * @param Generic $session
     * @param CustomerSession $customerSession
     * @param ResolverInterface $localeResolver
     * @param UrlInterface $urlBuilder
     * @param TruevoConfig $truevoConfig
     * @param CoreHelper $coreHelper
     * @param array $data
     */
    public function __construct(
        Context $context,
        Registry $registry,
        ExtensionAttributesFactory $extensionFactory,
        AttributeValueFactory $customAttributeFactory,
        StoreManagerInterface $storeManager,
        Data $paymentData,
        ScopeConfigInterface $scopeConfig,
        Logger $logger,
        Session $checkoutSession,
        Generic $session,
        CustomerSession $customerSession,
        ResolverInterface $localeResolver,
        UrlInterface $urlBuilder,
        TruevoConfig $truevoConfig,
        CoreHelper $coreHelper,
        PaymentTokenManagementInterface $tokenManagement,
        array $data = array()
    ) {
        parent::__construct(
            $context,
            $registry,
            $extensionFactory,
            $customAttributeFactory,
            $paymentData,
            $scopeConfig,
            $logger,
            null,
            null,
            $data
        );

        $this->_storeManager    = $storeManager;
        $this->_checkoutSession = $checkoutSession;
        $this->_session         = $session;
        $this->_customerSession = $customerSession;
        $this->_paymentData     = $paymentData;
        $this->_localeResolver  = $localeResolver;
        $this->_urlBuilder      = $urlBuilder;
        $this->_truevoConfig    = $truevoConfig;
        $this->_coreHelper      = $coreHelper;
        $this->_tokenManagement = $tokenManagement;
    }

    public function getsupportedCurrencyCodes()
    {
        return explode(",", $this->getConfigData('allowed_currency'));
    }

    public function canUseForCurrency($currencyCode)
    {
        if (!in_array($currencyCode, $this->getsupportedCurrencyCodes())) {
            return false;
        }
        return true;
    }

    /**
     * Checks if the customer is logged in and tokenization is enabled in the configuration.
     *
     * @return boolean
     */
    private function canUseTokenization()
    {
        return $this->_customerSession->isLoggedIn() && $this->_truevoConfig->isTokenizationEnabled();
    }

    /**
     * Generate a unique shopper reference for tokenization.
     *
     * @return string
     */
    private function getCustomerTokenizationReference()
    {
        $customer = $this->_customerSession->getCustomer();

        if ($customer) {
            return "tokenization-" . $customer->getId() . '-' . $customer->getEmail() . '-' . $customer->getCreatedAtTimestamp();
        }

        // Fallback in case this function is hit without a signed in customer
        return null;
    }

    protected function prepareArgs($order)
    {
        //SET
        $billingAddress = $order->getBillingAddress();

        //PROCESSING
        $args = array(
            'transaction_type'    => $this->_truevoConfig->getTransactionType(),
            'merchant_id'         => $this->_truevoConfig->getMid(),
            'terminal_id'         => $this->_truevoConfig->getTid(),
            'language'            => $this->_localeResolver->getLocale(),
            'show_logo'           => $this->_truevoConfig->showLogo(),
            'widget_theme'        => $this->_truevoConfig->getPaymentStyle(),
            'css_styles'          => $this->minifyCss($this->_truevoConfig->getcss()),
            'payment_reference'   => $this->_storeManager->getStore()->getName() . ' Purchase',
            'order_reference'     => 'Order #: ' . $order->getId() . ' | ' . date(DATE_W3C),
            'currency'            => $order->getBaseCurrencyCode(),
            'amount'              => doubleval($order->getBaseGrandTotal()),
            'gratuity_amount'     => 0,
            'transaction_source'  => 'ECOMMERCE',
            'customer_first_name' => $billingAddress->getFirstname(),
            'customer_last_name'  => $billingAddress->getLastname(),
            'customer_email'      => $billingAddress->getEmail(),
            'customer_phone'      => $this->sanitizePhoneNumber($billingAddress->getTelephone()),
            'billing_address'     => isset($billingAddress->getStreet()[0]) ? $billingAddress->getStreet()[0] : '',
            'billing_address_2'   => isset($billingAddress->getStreet()[1]) ? $billingAddress->getStreet()[1] : '',
            'billing_city'        => $billingAddress->getCity(),
            'billing_state'       => $billingAddress->getRegion(),
            'billing_country'     => $billingAddress->getCountryId(),
            'billing_postcode'    => $billingAddress->getPostCode(),
            'ip_address'          => $this->_coreHelper->getClientIpAddress(),
            'additional_data'     => array(),
        );

        // If tokenization is enabled and customer is logged in append widget configuration
        if ($this->canUseTokenization()) {
            $args['tokenization_shopper_reference']    = $this->getCustomerTokenizationReference();
            $args['tokenization_registration_process'] = 'Optional';
            $args['tokenization_type']                 = 'OneClickPayment';
        }

        //RETURN
        return $args;
    }

    protected function sanitizePhoneNumber($number)
    {
        //SET
        $return = "";

        //PROCESS
        $return = str_replace('+', '', $number);

        //RETURN
        return $return;
    }

    protected function getWidgetConfiguration($order)
    {
        $args = $this->prepareArgs($order);

        $this->_coreHelper->loggingTruevo(str_replace("\n", "", print_r($args, true)));

        $configuration = [
            "transactionType"     => $args['transaction_type'],
            "widgetConfiguration" => [
                "language"     => $args['language'],
                "showTruevo"   => $args['show_logo'],
                "iframeStyles" => [
                    "theme" => $args['widget_theme'],
                    "style" => $args['css_styles'],
                ],
            ],
            "transaction"         => [
                "paymentReference"  => $args['payment_reference'],
                "merchantReference" => $args['order_reference'],
                "amount"            => [
                    "totalAmount"       => $args['amount'],
                    "currencyAlphaCode" => $args['currency'],
                    "details"           => [
                        "paymentAmount"  => $args['amount'],
                        "gratuityAmount" => $args['gratuity_amount'],
                    ],
                ],
                "source"            => $args['transaction_source'],
            ],
            "merchant"            => [
                "mid" => $args['merchant_id'],
                "tid" => $args['terminal_id'],
            ],
            "customer"            => [
                "firstName"      => $args['customer_first_name'],
                "lastName"       => $args['customer_last_name'],
                "contact"        => [
                    "email" => $args['customer_email'],
                    "phone" => $args['customer_phone'],
                    "ip"    => $args['ip_address'],
                ],
                "billingAddress" => [
                    "street1"          => $args['billing_address'],
                    "street2"          => $args['billing_address_2'],
                    "city"             => $args['billing_city'],
                    "state"            => $args['billing_state'],
                    "postCodeZip"      => $args['billing_postcode'],
                    "countryAlphaCode" => $args['billing_country'],
                ],
            ],
            "additionalData"      => $args['additional_data'],
        ];

        // If tokenization attributes exist, pass them through to the configuration
        if (isset($args['tokenization_shopper_reference']) && isset($args['tokenization_registration_process']) && isset($args['tokenization_type'])) {
            $configuration['tokenization'] = [
                "merchantShopperReference" => $args['tokenization_shopper_reference'],
                "tokenRegistrationProcess" => $args['tokenization_registration_process'],
                "tokenType"                => $args['tokenization_type'],
            ];
        }

        $this->_coreHelper->_serializePrecisionFix();

        return json_encode($configuration);
    }

    public function initialize($paymentAction, $stateObject)
    {
        $payment = $this->getInfoInstance();
        $order   = $payment->getOrder();
        $order->setCanSendNewEmailFlag(false);

        $stateObject->setState(\Magento\Sales\Model\Order::STATE_PENDING_PAYMENT);
        $stateObject->setStatus('pending_payment');
        $stateObject->setIsNotified(false);
    }

    public function getPostHTML($order, $storeId = null)
    {
        $widgetHtml = "<div id='truevo_form_container'></div>
            <form class='truevo-cancel-form' action='" . $this->getOrderCancelledRedirectUrl() . "'>
                <button type='submit' id='truevo_cancel_button' class='truevo-cancel-button'>Cancel</button>
            </form>
            <script src='" . $this->_truevoConfig->getWidgetUrl() . "'></script>
            <script type='text/javascript'>
                const configuration = JSON.parse('" . $this->getWidgetConfiguration($order) . "');

                openCheckout();
                function openCheckout() {
                    TruevoCheckout.initialize(configuration);
                    const truevoCheckout = TruevoCheckout.mount('#truevo_form_container');
                    truevoCheckout.addEventListener('loadCompleted', res => {
                        console.log('======== Payment Widget Loaded ==========');
                    });
                    truevoCheckout.addEventListener('loadFailed', res => {
                        console.log('======== Payment Widget Failed To Load ==========');
                    });
                    truevoCheckout.addEventListener('transactionCompleted', res => {
                        let params = buildTransactionInfoParams(res.detail.transactionInfo, configuration)
                        postToCallback(params);
                    });
                    truevoCheckout.addEventListener('transactionDeclined', res => {
                        let params = buildTransactionInfoParams(res.detail.transactionInfo, configuration)
                        postToCallback(params);
                    });
                    truevoCheckout.addEventListener('transactionFailed', res => {
                        let params = buildTransactionInfoParams(res.detail.transactionInfo, configuration)
                        postToCallback(params);
                    });
                    truevoCheckout.addEventListener('payButtonClicked', res => {
                        document.getElementById('truevo_cancel_button').disabled = true;
                    });
                }

                function buildTransactionInfoParams(transactionInfo, widgetConfiguration) {
                    if (typeof transactionInfo === 'undefined') {
                        return {
                            transactionType: widgetConfiguration.transactionType,
                            rrn: null,
                            transactionId: null,
                            transactionStatus: null,
                            responseCode: null,
                            reasonDescription: null,
                            merchantReference: widgetConfiguration.transaction.merchantReference
                        }
                    }
                    return {
                        transactionType: widgetConfiguration.transactionType,
                        rrn: transactionInfo.rrn,
                        transactionId: transactionInfo.transactionId,
                        transactionStatus: transactionInfo.transactionStatus,
                        responseCode: transactionInfo.responseCode,
                        reasonDescription: typeof transactionInfo.reasonDescription === 'undefined' ? null : transactionInfo.reasonDescription,
                        merchantReference: transactionInfo.merchantReference
                    }
                }

                function postToCallback(params) {
                    const form = document.createElement('form');
                    form.method = 'POST';
                    form.action = '" . $this->getOrderCallbackUrl() . "';
                    for (const key in params) {
                        if (params.hasOwnProperty(key)) {
                            const hiddenField = document.createElement('input');
                            hiddenField.type = 'hidden';
                            hiddenField.name = key;
                            hiddenField.value = params[key];
                            form.appendChild(hiddenField);
                        }
                    }
                    document.body.appendChild(form);
                    form.submit();
                }
            </script>";

        return $widgetHtml;
    }

    public function getOrderPlaceRedirectUrl($storeId = null)
    {
        return $this->_getUrl('truevo/checkout', $storeId);
    }

    public function getOrderCallbackUrl($storeId = null)
    {
        return $this->_getUrl('truevo/callback', $storeId);
    }

    public function getOrderCancelledRedirectUrl($storeId = null)
    {
        return $this->_getUrl('truevo/cancel', $storeId);
    }

    public function getSuccessUrl($storeId = null)
    {
        return $this->_getUrl('checkout/onepage/success', $storeId);
    }

    public function getCancelUrl($storeId = null)
    {
        return $this->_getUrl('checkout/onepage/failure', $storeId);
    }

    public function getCheckout()
    {
        return $this->_checkoutSession;
    }

    public function getQuote()
    {
        return $this->getCheckout()->getQuote();
    }

    protected function _getUrl($path, $storeId, $secure = null)
    {
        $store = $this->_storeManager->getStore($storeId);

        return $this->_urlBuilder->getUrl(
            $path,
            ['_store' => $store, '_secure' => $secure === null ? $store->isCurrentlySecure() : $secure]
        );
    }

    /**
     * @return void
     */
    public function validate()
    {
        $data = $this->getInfoInstance();

        if ($data instanceof \Magento\Sales\Model\Order\Payment) {
            $billing_Country = $data->getOrder()->getBillingAddress()->getCountryId();
            $currency_code   = $data->getOrder()->getBaseCurrencyCode();
        } else {
            $billing_Country = $data->getQuote()->getBillingAddress()->getCountryId();
            $currency_code   = $data->getQuote()->getBaseCurrencyCode();
        }

        if (!$this->canUseForCountry($billing_Country)) {
            throw new \Magento\Framework\Validator\Exception(__('Selected payment type is not allowed for billing country.'));
        }

        if (isset($currency_code) && $currency_code != '' && !in_array($currency_code, $this->getsupportedCurrencyCodes())) {
            throw new \Magento\Framework\Exception\LocalizedException(
                __('Selected currency code (' . $currency_code . ') is not compatible.')
            );
        }

        return $this;
    }

    protected function minifyCss($css)
    {
        return str_replace(array("\r", "\n"), '', $css);
    }
}
