<?php

namespace Truevo\PaymentGateway\Model;

use Magento\Framework\App\ObjectManager;
use Magento\Sales\Model\Order;
use Magento\Store\Model\StoreManagerInterface;
use Truevo\PaymentGateway\Model\TruevoConfig;
use Truevo\PaymentGateway\Helper\Data as CoreHelper;
use Truevo\PaymentGateway\Logger\Logger;

class TruevoAPI
{
    /**
     * The root module.
     *
     * @var TruevoConfig
     */
    private $_truevoConfig;

    /**
     * A unique value generated to recognize subsequent retries of the same request.
     *
     * @var string
     */
    private $_idempotencyKey;

    /**
     * The order saved in the store to match up with Truevo transaction
     *
     * @var Order
     */
    private $_order;

    /**
     * The order saved in the store to match up with Truevo transaction
     *
     * @var StoreManagerInterface
     */
    private $_storeManager;

    /**
     * @var Logger
     */
    private $_truevologger;

    /**
     * The Data helper file
     *
     * @var CoreHelper
     */
    private $_coreHelper;

    /**
     * Constructor
     *
     * @param Order $order
     */
    public function __construct(
        Order $order
    )
    {
        $objectManager = ObjectManager::getInstance();

        $this->_idempotencyKey = $this->generateUuid();
        $this->_order = $order;
        $this->_truevoConfig = $objectManager->get(TruevoConfig::class);
        $this->_storeManager = $objectManager->get(StoreManagerInterface::class);
        $this->_truevologger = $objectManager->get(Logger::class);
        $this->_coreHelper   = $objectManager->get(CoreHelper::class);
    }

    /**
     * Generate a valid RFC 4211 Compliant version 4 Universally Unique Identifier.
     *
     * @return string
     */
    private function generateUuid() 
    {
        return sprintf( '%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
            // 32 bits for "time_low"
            mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ),

            // 16 bits for "time_mid"
            mt_rand( 0, 0xffff ),

            // 16 bits for "time_hi_and_version",
            // four most significant bits holds version number 4
            mt_rand( 0, 0x0fff ) | 0x4000,

            // 16 bits, 8 bits for "clk_seq_hi_res",
            // 8 bits for "clk_seq_low",
            // two most significant bits holds zero and one for variant DCE1.1
            mt_rand( 0, 0x3fff ) | 0x8000,

            // 48 bits for "node"
            mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff )
        );
    }

    /**
     * Get the details of a transaction for refunds and reversals.
     *
     * @return string|false
     */
    public function getDetails() 
    {
        $transactionDetails = $this->getTransactionResponseDetails();

        if (!$transactionDetails) {
            $this->_coreHelper->loggingTruevo('Failed to find Truevo transaction details for capturing payment on Order #' . $this->_order->getId());
            return false;
        }

        $data = [
            'transactionUniqueIdentifier' => $transactionDetails['transactionId']
        ];

        $response = $this->sendRequest('GET', 'payments/details', $data);

        // Build missing properties
        if ($response && isset($response->paymentDetails)) {
            if (!property_exists($response->paymentDetails, 'isCaptureAllowed')) $response->paymentDetails->isCaptureAllowed = false;
            if (!property_exists($response->paymentDetails, 'isRefundAllowed')) $response->paymentDetails->isRefundAllowed = false;
            if (!property_exists($response->paymentDetails, 'isReverseAllowed')) $response->paymentDetails->isReverseAllowed = false;
            if (!property_exists($response->paymentDetails, 'availableForCapture')) $response->paymentDetails->availableForCapture = 0;
            if (!property_exists($response->paymentDetails, 'availableForRefund')) $response->paymentDetails->availableForRefund = 0;

            return $response;
        }

        return false;
    }

    /**
     * Get the status of a transaction (to be used after a capture payment call).
     *
     * @param string $transactionUniqueIdentifier The unique identifier to check (A capture request counts as a different transaction and will return a new identifier).
     * @return string|false
     */
    public function getStatus($transactionUniqueIdentifier)
    {
        $data = [
            'transactionUniqueIdentifier' => $transactionUniqueIdentifier,
            'merchantIdentifier'          => $this->_truevoConfig->getMid(),
            'terminalIdentifier'          => $this->_truevoConfig->getTid()
        ];

        return $this->sendRequest('GET', 'payments/status', $data);
    }

    /**
     * Capture a payment for an order currently "On Hold".
     *
     * @return string|false
     */
    public function capturePayment()
    {
        $transactionResponseDetails = $this->getTransactionResponseDetails();

        if (is_null($transactionResponseDetails)) {
            $this->_coreHelper->loggingTruevo('Failed to find Truevo transaction details for capturing payment on Order #' . $this->_order->getId());
            return false;
        }

        $billingAddress = $this->_order->getBillingAddress();

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionResponseDetails['transactionId'],
                'retrievalReferenceNumber' => $transactionResponseDetails['rrn'],
                "paymentReference" => $this->_storeManager->getStore()->getName() . ' Purchase',
                "merchantReference" => $transactionResponseDetails['merchantReference'],
                "amount" => [
                    "totalAmount" => (float) $this->_order->getGrandTotal(),
                    "currencyAlphaCode" => $this->_order->getBaseCurrencyCode()
                ],
                "source" => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->_truevoConfig->getMid(),
                "tid" => $this->_truevoConfig->getTid(),
            ],
            "customer" => [
                "firstName" => $billingAddress->getFirstname(),
                "lastName" => $billingAddress->getLastname(),
                "contact" => [
                    "email" => $billingAddress->getEmail(),
                    "phone" => $billingAddress->getTelephone(),
                    "ip" => $this->_coreHelper->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => isset($billingAddress->getStreet()[0]) ? $billingAddress->getStreet()[0] : '',
                    "street2" => isset($billingAddress->getStreet()[1]) ? $billingAddress->getStreet()[1] : '',
                    "city" => $billingAddress->getCity(),
                    "postCodeZip" => $billingAddress->getPostcode(),
                    "state" => $billingAddress->getRegion(),
                    "countryAlphaCode" => $billingAddress->getCountryId(),
                ]
            ],
            "additionalData" => []
        ];

        return $this->sendRequest('POST', 'payments/capture', $data);
    }

    /**
     * Send through a partial refund or a full refund request for an order.
     *
     * @param float|null $amount The amount to refund.
     * @param integer|null $numberOfPieces The number of products being returned.
     * @param string|null $reason The reason for the refund.
     * @return string|false
     */
    public function refund($amount = null, $numberOfPieces = null, $reason = null)
    {
        $transactionResponseDetails = $this->getTransactionResponseDetails();

        if (is_null($transactionResponseDetails)) {
            $this->_coreHelper->loggingTruevo('Failed to find Truevo transaction details for capturing payment on Order #' . $this->_order->getId());
            return false;
        }
        
        $additionalData = array();

        if (!is_null($numberOfPieces)) {
            $additionalData[] = [
                'key' => 'number_of_pieces',
                'value' => $numberOfPieces
            ];
        }

        if (!is_null($reason) && !empty($reason)) {
            $additionalData[] = [
                'key' => 'reason_for_refund',
                'value' => $reason
            ];
        }
        
        $totalAmount = !is_null($amount) ? $amount : $this->_order->getGrandTotal();

        $billingAddress = $this->_order->getBillingAddress();

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionResponseDetails['transactionId'],
                'retrievalReferenceNumber' => $transactionResponseDetails['rrn'],
                'paymentReference' => $this->_storeManager->getStore()->getName() . ' Purchase',
                'merchantReference' => $transactionResponseDetails['merchantReference'],
                'amount' => [
                    'totalAmount' => (float) $totalAmount,
                    'currencyAlphaCode' => $this->_order->getBaseCurrencyCode()
                ],
                'source' => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->_truevoConfig->getMid(),
                "tid" => $this->_truevoConfig->getTid(),
            ],
            "customer" => [
                "firstName" => $billingAddress->getFirstname(),
                "lastName" => $billingAddress->getLastname(),
                "contact" => [
                    "email" => $billingAddress->getEmail(),
                    "phone" => $billingAddress->getTelephone(),
                    "ip" => $this->_coreHelper->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => isset($billingAddress->getStreet()[0]) ? $billingAddress->getStreet()[0] : '',
                    "street2" => isset($billingAddress->getStreet()[1]) ? $billingAddress->getStreet()[1] : '',
                    "city" => $billingAddress->getCity(),
                    "postCodeZip" => $billingAddress->getPostcode(),
                    "state" => $billingAddress->getRegion(),
                    "countryAlphaCode" => $billingAddress->getCountryId(),
                ]
            ],
            "additionalData" => $additionalData
        ];

        return $this->sendRequest('POST', 'payments/linkedRefund', $data);
    }

    /**
     * Void the transaction and reverse it's effects. 
     * This operation can only be performed until a cut-off point for the transaction.
     *
     * @return string|false
     */
    public function reversal()
    {
        $transactionResponseDetails = $this->getTransactionResponseDetails();

        if (is_null($transactionResponseDetails)) {
            $this->_coreHelper->loggingTruevo('Failed to find Truevo transaction details for capturing payment on Order #' . $this->_order->getId());
            return false;
        }
   
        $billingAddress = $this->_order->getBillingAddress();

        $data = [
            "transaction" => [
                'originalTransactionUniqueIdentifier' => $transactionResponseDetails['transactionId'],
                'retrievalReferenceNumber' => $transactionResponseDetails['rrn'],
                'paymentReference' => $this->_storeManager->getStore()->getName() . ' Purchase',
                'merchantReference' => $transactionResponseDetails['merchantReference'],
                'amount' => [
                    'totalAmount' => (float) $this->_order->getGrandTotal(),
                    'currencyAlphaCode' => $this->_order->getBaseCurrencyCode()
                ],
                'source' => 'ECOMMERCE'
            ],
            "merchant" => [
                "mid" => $this->_truevoConfig->getMid(),
                "tid" => $this->_truevoConfig->getTid(),
            ],
            "customer" => [
                "firstName" => $billingAddress->getFirstname(),
                "lastName" => $billingAddress->getLastname(),
                "contact" => [
                    "email" => $billingAddress->getEmail(),
                    "phone" => $billingAddress->getTelephone(),
                    "ip" => $this->_coreHelper->getClientIpAddress(),
                ],
                "billingAddress" => [
                    "street1" => isset($billingAddress->getStreet()[0]) ? $billingAddress->getStreet()[0] : '',
                    "street2" => isset($billingAddress->getStreet()[1]) ? $billingAddress->getStreet()[1] : '',
                    "city" => $billingAddress->getCity(),
                    "postCodeZip" => $billingAddress->getPostcode(),
                    "state" => $billingAddress->getRegion(),
                    "countryAlphaCode" => $billingAddress->getCountryId(),
                ]
            ],
            "additionalData" => []
        ];

        return $this->sendRequest('POST', 'payments/reversal', $data);
    }

    /**
     * Sends the cURL request to the API.
     *
     * @param string $method POST | GET
     * @param string $endpoint The endpoint to send to.
     * @param array $data The values to send to the API
     * @return string
     */
    private function sendRequest($method, $endpoint, $data) 
    {
        $headers = array(
            "Authorization: Bearer " . $this->_truevoConfig->getAuthToken(),
            "Accept: application/json",
            "Content-Type: application/json; charset=utf-8"
        );

        $queryParams = $method === "GET" ? http_build_query($data) : null;
        $url = $this->_truevoConfig->getFullTruevoEndpoint($endpoint, $queryParams);

        $this->_coreHelper->loggingTruevo('Truevo payment sending cURL request to ' . $url . ' with the following data: ' . print_r($data, true));
        $this->_coreHelper->_serializePrecisionFix();

        $curl = curl_init();

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_ENCODING, "");
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 30);
        curl_setopt($curl, CURLOPT_TIMEOUT, 30);

        if ($method === "POST") {
            $headers[] = "IdempotencyKey: " . $this->_idempotencyKey;

            curl_setopt($curl, CURLOPT_POST, 1);
            curl_setopt($curl, CURLOPT_POSTFIELDS, json_encode($data));
        }

        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);

        $response = curl_exec($curl);

        if (curl_errno($curl)) {
            $this->_coreHelper->loggingTruevo('Truevo payment error sending cURL request: ' . print_r(curl_error($curl), true));

            return false;
        }

        curl_close($curl);

        return json_decode($response);
    }

    /**
     * Get the transaction response details from the order's status history.
     *
     * @return array
     */
    public function getTransactionResponseDetails()
    {
        $result = null;

        $histories = $this->_order->getStatusHistories();

        foreach ($histories as $key => $history) {
            $comment = $history->getComment();

            // Check if the comment contains the transaction response details
            if (strpos($comment, 'Transaction Type:') === 0) {
                $result = [
                    'transactionType'   => $this->_coreHelper->getStringBetween($comment, 'Transaction Type:', '<'),
                    'rrn'               => $this->_coreHelper->getStringBetween($comment, 'RRN:', '<'),
                    'transactionId'     => $this->_coreHelper->getStringBetween($comment, 'Transaction ID:', '<'),
                    'transactionStatus' => $this->_coreHelper->getStringBetween($comment, 'Transaction status:', '<'),
                    'responseCode'      => $this->_coreHelper->getStringBetween($comment, 'Response Code:', '<'),
                    'merchantReference' => $this->_coreHelper->getStringBetween($comment, 'Merchant Reference:', '<'),
                ];

                break;
            }
        }

        // If the order was a pre-auth check for the new Transaction ID after a capture
        if ($this->_checkIfPreauth($result)) {
            $captureTransactionId = $this->_getCaptureTransactionId($histories);

            // Update the Transaction ID with the new one
            if ($captureTransactionId) {
                $result['originalTransactionId'] = $result['transactionId'];
                $result['transactionId'] = $captureTransactionId;
            }
        }

        return $result;
    }

    /**
     * Check if the transaction type was "AUTHORIZE".
     * 
     * @param array $transactionDetails
     * @return boolean
     */
    private function _checkIfPreauth($transactionDetails)
    {
        if (!is_array($transactionDetails)) {
            return false;
        }

        return $transactionDetails['transactionType'] === "AUTHORIZE";
    }

    /**
     * Get the Transaction ID from a captured payment.
     * 
     * @param HistoryCollection $histories
     * @return string|null
     */
    private function _getCaptureTransactionId($histories)
    {
        foreach ($histories as $key => $history) {
            $comment = $history->getComment();

            if (strpos($comment, 'Action: CAPTURE') === 0) {
                return $this->_coreHelper->getStringBetween($comment, 'Transaction Unique Identifier:', '<');
            }
        }

        return null;
    }
}
